import enum
import time
import threading
import logging


log = logging.getLogger(__name__)


class State(enum.Enum):
    STOPPED = 'stopped'
    STARTING = 'starting'
    RUNNING = 'running'
    PAUSING = 'pausing'
    PREPARING = 'preparing'
    PAUSED = 'paused'
    STOPPING = 'stopping'
    EXCEPTION = 'exception'


STOPPED = State.STOPPED
STARTING = State.STARTING
RUNNING = State.RUNNING
PAUSING = State.PAUSING
PREPARING = State.PREPARING
PAUSED = State.PAUSED
STOPPING = State.STOPPING
EXCEPTION = State.EXCEPTION


intermediate_states = {
    RUNNING: PAUSED,
    STOPPED: PAUSED,
}


class Service:

    State = globals()['State']

    def __init__(self, name, worker):
        self.name = name
        self.worker = worker
        self.exception = None
        self.state_listeners = set()
        self.next_transition = None
        self.state_lock = threading.RLock()
        self._target_state = None
        self._next_state = None
        self._state = STOPPED
        self._transition = None
        self.state_timestamp = time.time()
        worker.service = self

    def register_state_change_listener(self, callback):
        self.state_listeners.add(callback)

    def unregister_state_change_listener(self, callback):
        self.state_listeners.discard(callback)

    def start(self):
        self._transition_to(RUNNING)

    def pause(self):
        self._transition_to(PAUSED)

    def stop(self):
        self._transition_to(STOPPED)

    prepare = pause

    def _transition_to(self, target_state):
        with self.state_lock:
            if self._state == EXCEPTION:
                log.debug('_transition_to(%s) -> EXCEPTION' % target_state)
                return
            if self._state == target_state:
                self._target_state = None
                self._next_state = None
                log.debug('_transition_to(%s) -> NOP' % target_state)
                return
            if (self._target_state == target_state and self._transition and
                    self._transition[1] == target_state):
                # already transitioning to give state
                self._target_state = None
                self._next_state = None
                log.debug('_transition_to(%s) -> in progress' % target_state)
                return
            transition = (self._state, target_state)
            if transition in self.worker._state_transitions:
                log.debug('_transition_to(%s) -> transition initiated' %
                          target_state)
                self._target_state = None
                self._next_state = None
                if target_state == RUNNING:
                    self.state = STARTING
                elif target_state == STOPPED:
                    self.state = STOPPING
                elif target_state == PAUSED and self.state == STOPPED:
                    self.state = PREPARING
                elif target_state == PAUSED:
                    self.state = PAUSING
                self._target_state = target_state
                funcname = self.worker._state_transitions[transition]
                callback = getattr(self.worker, funcname)
                self._transition = transition
                threading.Thread(target=self._execute_transition,
                                 args=(transition, callback)).start()
                return
            if target_state in intermediate_states:
                log.debug('_transition_to(%s) -> intermediate(%s)' % (
                    target_state, intermediate_states[target_state]))
                self._transition_to(intermediate_states[target_state])
                self._next_state = target_state
            else:
                log.debug('_transition_to(%s) -> queued' % target_state)
                self._next_state = target_state

    def _execute_transition(self, transition, callback):
        log.debug('_execute_transition(%s, %s)' % (str(transition),
                                                   callback.__name__))
        state_timestamp = self.state_timestamp
        try:
            callback()
            with self.state_lock:
                if self._transition == transition:
                    self._transition = None
                if state_timestamp >= self.state_timestamp:
                    self.state = transition[1]
        except Exception as exception:
            self.set_exception(exception)

    def set_exception(self, exception):
        with self.state_lock:
            if self._state == EXCEPTION:
                return
            old_state = self._state
            self._state = EXCEPTION
            self.exception = exception
            self.state_timestamp = time.time()
        self.worker.cleanup(exception)
        self._state_changed(old_state, EXCEPTION)

    def _state_changed(self, old, new):
        log.debug('changed state: %s -> %s' % (old, new))
        if new == EXCEPTION:
            log.exception(self.exception)
        for callback in self.state_listeners:
            callback(self, old, new)
        with self.state_lock:
            if self._next_state:
                log.debug('  next state: %s' % (self._next_state))
                next_state = self._next_state
                self._next_state = None
                self._transition_to(next_state)
            elif self._target_state and self._state != self._target_state:
                log.debug('  target state: %s' % (self._target_state))
                self._transition_to(self._target_state)

    @property
    def state(self):
        return self._state

    @state.setter
    def state(self, new_state):
        with self.state_lock:
            if new_state == self._state or self._state == EXCEPTION:
                return
            old_state = self._state
            self._state = new_state
            self.state_timestamp = time.time()
        self._state_changed(old_state, new_state)



import asyncio
import multiprocessing
import threading
import os
import functools
import signal
import sys
from tblib import pickling_support


pickling_support.install()


def fork(loop, cls, *args, **kwargs):
    parent_pipe, child_pipe = multiprocessing.Pipe()
    assert threading.active_count() == 1, "Cannot fork when using threads"
    childpid = os.fork()
    if childpid:
        return Gateway(loop, cls, childpid, parent_pipe)
    signal.signal(signal.SIGINT, signal.SIG_IGN)
    obj = cls(*args, **kwargs)
    obj.pipe = child_pipe
    loop = asyncio.get_event_loop()
    if loop.is_running():
        loop.stop()
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
    loop.add_reader(obj.pipe.fileno(), _handle_call, obj)
    loop.run_forever()
    os._exit(0)


def _handle_call(obj):
    def done(future):
        exc = future.exception()
        if exc:
            obj.pipe.send((id, False, (type(exc), exc, exc.__traceback__)))
        else:
            obj.pipe.send((id, True, future.result()))
    command = obj.pipe.recv()
    id, funcname, args, kwargs = command
    try:
        if funcname == '_get_attribute':
            result = getattr(obj, args[0])
        else:
            callback = getattr(obj, funcname)
            result = callback(*args, **kwargs)
            if asyncio.iscoroutine(result):
                loop = asyncio.get_event_loop()
                loop.create_task(result).add_done_callback(done)
                return
        obj.pipe.send((id, True, result))
    except:
        obj.pipe.send((id, False, sys.exc_info()))


class Backgrounded:

    def trigger(self, event, *args):
        self.pipe.send((event, args))

    def kill(self):
        loop = asyncio.get_event_loop()
        loop.stop()


class Gateway:

    def __init__(self, loop, cls, childpid, pipe):
        self.cls = cls
        self.childpid = childpid
        self.pipe = pipe
        self.last_command_id = 0
        self.responses = {}
        self.loop = loop
        self.loop.add_reader(self.pipe.fileno(), self._message_received)
        self.event = asyncio.Event(loop=loop)
        self.callbacks = {}

    def on(self, event, callback):
        if event not in self.callbacks:
            self.callbacks[event] = []
        self.callbacks[event].append(callback)

    def off(self, event, callback):
        self.callbacks[event].remove(callback)
        if not self.callbacks[event]:
            del self.callbacks[event]

    def _message_received(self):
        try:
            message = self.pipe.recv()
        except EOFError:
            return
        if len(message) == 2:
            event, args = message
            if event not in self.callbacks:
                return
            for callback in self.callbacks[event]:
                result = callback(*args)
                if asyncio.iscoroutine(result):
                    self.loop.create_task(result)
        else:
            id, success, result = message
            self.responses[id] = (success, result)
            self.event.set()

    @asyncio.coroutine
    def kill(self):
        if not self.childpid:
            return
        try:
            yield from self._send_command('kill')
        except BrokenPipeError:
            pass
        if not self.childpid:
            return
        os.waitpid(self.childpid, 0)
        self.childpid = None

    def __del__(self):
        if self.childpid:
            os.kill(self.childpid, signal.SIGTERM)

    def __getattr__(self, name):
        if name.startswith('_'):
            raise AttributeError('Cannot access protected member %s.%s' %
                                 (self.cls.__name__, name))
        member = getattr(self.cls, name)
        if not callable(member):
            return self._send_command('_get_attribute', name)

        @functools.wraps(member)
        def callback(*args, **kwargs):
            return self._send_command(name, *args, **kwargs)
        setattr(self, name, callback)
        return callback

    @asyncio.coroutine
    def _send_command(self, command, *args, **kwargs):
        command_id = self.last_command_id + 1
        self.last_command_id += 1
        self.pipe.send((command_id, command, args, kwargs))
        while True:
            yield from self.event.wait()
            self.event.clear()
            if command_id in self.responses:
                success, result = self.responses[command_id]
                break
        if success:
            return result
        raise result[1].with_traceback(result[2])



import abc
from ..service import Service


STOPPED = Service.State.STOPPED
STARTING = Service.State.STARTING
RUNNING = Service.State.RUNNING
PAUSING = Service.State.PAUSING
PREPARING = Service.State.PREPARING
PAUSED = Service.State.PAUSED
STOPPING = Service.State.STOPPING
EXCEPTION = Service.State.EXCEPTION


final_states = {
    'start': RUNNING,
    'stop': STOPPED,
    'pause': PAUSED,
    'prepare': PAUSED,
}

invalid_transitions = {
    (STARTING, RUNNING),
    (STOPPING, STOPPED),
    (PAUSING, PAUSED),
    (PREPARING, PAUSED),
}


def transitions(state1, state2=None):
    def wrapper(func):
        nonlocal state1, state2
        if not hasattr(func, 'transitions'):
            func.transitions = set()
        name = func.__name__
        if not state2:
            try:
                state2 = final_states[name]
            except KeyError:
                raise Exception(
                    'Function %s() has no end state for transition from %s' %
                    (name, state1))
        elif name in final_states and final_states[name] != state2:
            raise Exception('Function %s() must transition to %s' %
                            (name, final_states[name]))
        for transition in func.transitions:
            if state2 != transition[1]:
                raise Exception(
                    'Function %s() cannot traansition to both %s and %s' %
                    (name, state2, transition[1]))
        func.transitions.add((state1, state2))
        return func
    if state1 == state2 or (state1, state2) in invalid_transitions:
        raise ValueError('Invalid transition: %s' % str((state1, state2)))
    return wrapper


class WorkerMeta(abc.ABCMeta):

    def __init__(cls, name, parents, members):
        transitions = {}
        for name in members:
            func = members[name]
            if not callable(func):
                continue
            if not hasattr(func, 'transitions'):
                continue
            for transition in func.transitions:
                if transition in transitions:
                    raise Exception(
                        'Transition %s already registered as function %s' %
                        (str(transition), transitions[transition]))
                transitions[transition] = name
        if not parents:
            tmp = {
                (STOPPED, PAUSED): 'prepare',
                (PAUSED, RUNNING): 'start',
                (RUNNING, PAUSED): 'pause',
                (PAUSED, STOPPED): 'stop',
            }
            tmp.update(transitions)
            transitions = tmp
        if transitions:
            for parent in parents:
                if hasattr(parent, '_state_transitions'):
                    tmp = parent._state_transitions.copy()
                    tmp.update(transitions)
                    transitions = tmp
                    break
            members['_state_transitions'] = transitions
            cls._state_transitions = transitions
        abc.ABCMeta.__init__(cls, name, parents, members)


class Worker(metaclass=WorkerMeta):

    @property
    def state(self):
        return self.service.state

    @abc.abstractmethod
    def prepare(self):
        # only called when in stopped state
        # assumed to be paused when finished
        pass

    @abc.abstractmethod
    def start(self):
        # only called when in paused state
        # assumed to be running when finished
        pass

    @abc.abstractmethod
    def stop(self):
        # can be called at any time, even during transitions
        # assumed to be stopped when finished
        pass

    @abc.abstractmethod
    def pause(self):
        # only called when in running state
        # assumed to be paused when finished
        pass

    @abc.abstractmethod
    def cleanup(self, exception):
        # only called when an exception occured
        pass



import abc
import functools
import select
import socket
import socketserver
import threading

from .worker import Worker, transitions
from ..service import Service


class SocketServerWorker(Worker):

    final_states = (
        Service.State.STOPPED, Service.State.STOPPING, Service.State.EXCEPTION)

    def __init__(self):
        self.__server = None
        self.__num_running = 0
        self.__request_lock = threading.Condition()

    def prepare(self):
        self.__intr_pair = socket.socketpair()
        server = self._mkserver()
        assert isinstance(server, socketserver.BaseServer)
        self.__server = server
        original_shutdown = server.shutdown_request

        @functools.wraps(server.shutdown_request)
        def shutdown_request(*args, **kwargs):
            original_shutdown(*args, **kwargs)
            with self.__request_lock:
                self.__num_running -= 1
                self.__request_lock.notify()

        server.shutdown_request = shutdown_request
        threading.Thread(target=self._loop).start()

    def start(self):
        self._interrupt_loop()

    @transitions(Service.State.PREPARING)
    @transitions(Service.State.RUNNING)
    def stop(self):
        self._interrupt_loop()

    def pause(self):
        self._interrupt_loop()

    def cleanup(self, exception):
        if self.__server:
            try:
                self.__server.server_close()
                self.__server = None
            except:
                pass

    def _loop(self):
        if not self.__intr_pair:
            return
        while self.state not in self.final_states:
            try:
                if self.state == Service.State.RUNNING:
                    sockets = (self.__server.socket, self.__intr_pair[0],)
                else:
                    sockets = (self.__intr_pair[0],)
                r, w, e = select.select(sockets, [], [])
                if self.__intr_pair[0] in r:
                    self.__intr_pair[0].recv(2**10)
                    continue
            except InterruptedError:
                continue
            self._process_request()
        self.__server.server_close()
        self.__server = None

    def _process_request(self):
        server = self.__server
        with self.__request_lock:
            if self.state in self.final_states:
                # we're stopping, abort operation
                return
            request, client_address = server.get_request()
            self.__num_running += 1
        if not server.verify_request(request, client_address):
            return
        try:
            server.process_request(request, client_address)
        except:
            server.handle_error(request, client_address)
            server.shutdown_request(request)

    def _interrupt_loop(self):
        with self.__request_lock:
            if not self.__intr_pair:
                return
            self.__intr_pair[1].send(b'0')
            while self.__num_running:
                self.__request_lock.wait()

    @abc.abstractmethod
    def _mkserver(self):
        pass



# Copyright © 2015,2016 STRG.AT GmbH, Vienna, Austria
#
# This file is part of the The SCORE Framework.
#
# The SCORE Framework and all its parts are free software: you can redistribute
# them and/or modify them under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation which is in the
# file named COPYING.LESSER.txt.
#
# The SCORE Framework and all its parts are distributed without any WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. For more details see the GNU Lesser General Public
# License.
#
# If you have not received a copy of the GNU Lesser General Public License see
# http://www.gnu.org/licenses/.
#
# The License-Agreement realised between you as Licensee and STRG.AT GmbH as
# Licenser including the issue of its valid conclusion and its pre- and
# post-contractual effects is governed by the laws of Austria. Any disputes
# concerning this License-Agreement including the issue of its valid conclusion
# and its pre- and post-contractual effects are exclusively decided by the
# competent court, in whose district STRG.AT GmbH has its registered seat, at
# the discretion of STRG.AT GmbH also the competent court, in whose district the
# Licensee has his registered seat, an establishment or assets.

import watchdog.events
import watchdog.observers
import watchdog.utils
import os
import sys
import logging
import threading
import time

log = logging.getLogger('score.serve.changedetector')


Observer = watchdog.observers.Observer

if Observer.__name__ == 'InotifyObserver':
    import watchdog.observers.inotify

    # The inotify original observer has a small delay for pairing IN_MOVED_FROM
    # and IN_MOVE_TO events. Since we do not care whether something was moved
    # *into* or *out of* our watches, we will override this behaviour to achieve
    # the fastest possible reload times.

    class InotifyBuffer(watchdog.observers.inotify.InotifyBuffer):

        def __init__(self, *args, **kwargs):
            self.delay = 0
            super().__init__(*args, **kwargs)

        def close(self):
            self.stop()

    class InotifyEmitter(watchdog.observers.inotify.InotifyEmitter):

        def on_thread_start(self):
            path = watchdog.utils.unicode_paths.encode(self.watch.path)
            self._inotify = InotifyBuffer(path, self.watch.is_recursive)

    class InotifyObserver(watchdog.observers.inotify.InotifyObserver):

        def __init__(self, *args, **kwargs):
            watchdog.observers.api.BaseObserver.__init__(
                self, *args, emitter_class=InotifyEmitter, **kwargs)

    Observer = InotifyObserver


class ChangeDetector(watchdog.events.FileSystemEventHandler):

    def __init__(self, *, autostart=True):
        self.callbacks = []
        self.observer = Observer()
        # set thread name
        self.observer.name = 'ChangeDetector'
        self.gatherer = threading.Thread(target=self._gather_modules)
        self.running = False
        self._observer_lock = threading.Lock()
        if autostart:
            self.start()

    def _gather_modules(self):
        while self.running:
            for module in list(sys.modules.values()):
                self.observe_module(module)
                if not self.running:
                    return
            time.sleep(0.5)

    def start(self):
        self.observed_files = set()
        self.observed_dirs = {}
        self.observed_modules = set()
        self.file2modules = {}
        self.running = True
        self.gatherer.start()
        with self._observer_lock:
            self.observer.start()

    def stop(self, wait=True):
        if not self.running:
            return
        self.running = False
        self.observer.stop()
        if wait:
            if threading.current_thread() != self.observer:
                self.observer.join()
            self.gatherer.join()

    def observe_module(self, module):
        if module in self.observed_modules:
            return
        self.observed_modules.add(module)
        try:
            file = module.__file__
        except AttributeError:
            return
        if not file:
            return
        old = None
        while not os.path.isfile(file):
            old = file
            file = os.path.dirname(file)
            if file == old:
                return
        if file[-4:] in ('.pyc', '.pyo'):
            file = file[:-1]
        self.observe(file, module)

    def observe(self, file, module=None):
        if not os.path.exists(file):
            return
        file = os.path.abspath(file)
        if file in self.observed_files:
            return
        self.observed_files.add(file)
        dir = os.path.dirname(file)
        if module:
            try:
                self.file2modules[file].add(module)
            except KeyError:
                self.file2modules[file] = {module}
        if dir in self.observed_dirs:
            return
        if self.running:
            # safeguarding against startup errors: self.observer.schedule fails
            # with errno EBADF (bad file descriptor) if the thread is not
            # running.  this happens most commonly when the Worker has a startup
            # issue, when lots of new watches are added.
            with self._observer_lock:
                for other in self.observed_dirs.copy():
                    if dir.startswith(other):
                        return
                    if other.startswith(dir):
                        log.debug('unscheduling %s in favor of %s' %
                                  (other, dir))
                        self.observer.unschedule(self.observed_dirs[other])
                        del self.observed_dirs[other]
                log.debug('scheduling %s' % (dir))
                self.observed_dirs[dir] = \
                    self.observer.schedule(self, dir, recursive=True)

    def add_callback(self, callback):
        self.callbacks.append(callback)
        return callback

    def remove_callback(self, callback):
        self.callbacks.remove(callback)
        return callback

    def clear_callbacks(self):
        self.callbacks = []

    def on_any_event(self, event):
        if isinstance(event, watchdog.events.DirCreatedEvent):
            return
        if isinstance(event, watchdog.events.DirModifiedEvent):
            return
        FileCreatedEvent = watchdog.events.FileCreatedEvent
        file = event.src_path
        if file in self.observed_files:
            try:
                modules = self.file2modules[file]
            except KeyError:
                modules = []
            log.debug('file changed: %s' % file)
        elif isinstance(event, FileCreatedEvent) and file.endswith('.py'):
            log.debug('new file: %s' % file)
            modules = []
        else:
            return
        for callback in self.callbacks:
            callback(file, modules)



# Copyright © 2015,2016 STRG.AT GmbH, Vienna, Austria
#
# This file is part of the The SCORE Framework.
#
# The SCORE Framework and all its parts are free software: you can redistribute
# them and/or modify them under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation which is in the
# file named COPYING.LESSER.txt.
#
# The SCORE Framework and all its parts are distributed without any WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. For more details see the GNU Lesser General Public
# License.
#
# If you have not received a copy of the GNU Lesser General Public License see
# http://www.gnu.org/licenses/.
#
# The License-Agreement realised between you as Licensee and STRG.AT GmbH as
# Licenser including the issue of its valid conclusion and its pre- and
# post-contractual effects is governed by the laws of Austria. Any disputes
# concerning this License-Agreement including the issue of its valid conclusion
# and its pre- and post-contractual effects are exclusively decided by the
# competent court, in whose district STRG.AT GmbH has its registered seat, at
# the discretion of STRG.AT GmbH also the competent court, in whose district the
# Licensee has his registered seat, an establishment or assets.

import asyncio
from score.init import (
    ConfiguredModule, parse_list, parse_bool, init_from_file,
    InitializationError)
from .service import Service
from ._forked import fork, Backgrounded
from ._changedetect import ChangeDetector
from collections import OrderedDict
from contextlib import contextmanager
import traceback
import signal
import logging

log = logging.getLogger('score.serve')


defaults = {
    'autoreload': False,
    'modules': []
}


def init(confdict):
    conf = defaults.copy()
    conf.update(confdict)
    modules = parse_list(conf['modules'])
    if not modules:
        import score.serve
        raise InitializationError(score.serve, 'No modules configured')
    autoreload = parse_bool(conf['autoreload'])
    return ConfiguredServeModule(conf['conf'], modules, autoreload)


class ConfiguredServeModule(ConfiguredModule):

    def __init__(self, conf, modules, autoreload):
        import score.serve
        ConfiguredModule.__init__(self, score.serve)
        self.conf = conf
        self.modules = modules
        self.autoreload = autoreload

    def start(self):
        reload = ServerInstance(self).reload
        while reload:
            log.info('reloading')
            reload = ServerInstance(self).reload


class ServerInstance:

    def __init__(self, conf):
        self.conf = conf
        self.loop = asyncio.new_event_loop()
        self.controller = fork(self.loop, ServiceController, self.conf)
        if self.conf.autoreload:
            self.controller.on('restart', self.restart)
        self.reload = False
        self.controller.on('state-change', self.quit_if_stopped)
        # self.loop.set_debug(True)
        self.loop.add_signal_handler(signal.SIGINT, self.signal_handler_stop)
        self.loop.create_task(self.controller.start()).add_done_callback(
            lambda *_: log.info('started'))
        self.loop.run_forever()
        self.loop.remove_signal_handler(signal.SIGINT)

    def signal_handler_stop(self):
        log.info('Ctrl+C detected, stopping')
        self.loop.remove_signal_handler(signal.SIGINT)
        self.loop.create_task(self.controller.stop())
        self.reload = False

    def quit_if_stopped(self, states):
        if not all(state in (Service.State.STOPPED, Service.State.EXCEPTION)
                   for state in states.values()):
            return
        self.controller.off('state-change', self.quit_if_stopped)
        self.stop_loop()

    @asyncio.coroutine
    def restart(self):
        self.reload = True
        yield from self.controller.stop()
        self.stop_loop()

    def stop_loop(self, *_):
        pending_tasks = [t for t in asyncio.Task.all_tasks(self.loop)
                         if not t.done()]
        if not pending_tasks:
            self.loop.stop()
        else:
            task = pending_tasks.pop()
            task.add_done_callback(self.stop_loop)


class ServiceController(Backgrounded):

    def __init__(self, conf):
        self.conf = conf
        self._services = None
        self._changedetector = None

    def start(self):
        if not self._services:
            self._init_services()
        self._call_on_subservices('start')

    def pause(self):
        if not self._services:
            self._init_services()
        self._call_on_subservices('pause')

    def stop(self):
        if not self._services:
            return
        if self._changedetector:
            self._changedetector.stop(wait=False)
            self._changedetector = None
        self._call_on_subservices('stop')

    def service_states(self):
        if not self._services:
            self._init_services()
        result = OrderedDict()
        for name, service in self._services.items():
            result[name] = service.state
        return result

    @property
    @contextmanager
    def _acquire_service_locks(self):
        acquired = []
        try:
            for service in self._services.values():
                service.state_lock.acquire()
                acquired.append(service)
            yield
        finally:
            for service in acquired:
                service.state_lock.release()

    def _init_services(self):
        if self.conf.autoreload:
            self._changedetector = ChangeDetector()
            self._changedetector.observe(self.conf.conf)
            self._changedetector.add_callback(self.restart)
        try:
            self._collect_services()
            for service in self._services.values():
                service.register_state_change_listener(
                    self._service_state_changed)
        except Exception as e:
            if not self._changedetector:
                raise

            self._services.clear()
            self.conf.log.exception(e)
            if isinstance(e, SyntaxError):
                self._changedetector.observe(e.filename)
            else:
                for frame in traceback.extract_tb(e.__traceback__):
                    self._changedetector.observe(frame[0])

    def _service_state_changed(self, service, old, new):
        states = {service.name: service.state
                  for service in self._services.values()}
        self.trigger('state-change', states)
        if new == Service.State.EXCEPTION:
            self.conf.log.exception(service.exception)

    def _collect_services(self):
        self._services = OrderedDict()
        score = init_from_file(self.conf.conf)
        if self._changedetector:
            for file in parse_list(score.conf['score.init']['_files']):
                self._changedetector.observe(file)
        for mod in self.conf.modules:
            workers = score._modules[mod].score_serve_workers()
            if isinstance(workers, list):
                for i, worker in enumerate(workers):
                    name = '%s/%d' % (mod, i)
                    self._services[name] = Service(name, worker)
            elif isinstance(workers, dict):
                for name, worker in workers.items():
                    name = '%s/%s' % (mod, name)
                    self._services[name] = Service(name, worker)

    def restart(self, *_):
        self.trigger('restart')
        changedetector = self._changedetector
        self._changedetector = None
        if changedetector:
            changedetector.stop(wait=False)
        self.stop()

    def _call_on_subservices(self, func, *args):
        for service in self._services.values():
            getattr(service, func)(*args)



# Copyright © 2015,2016 STRG.AT GmbH, Vienna, Austria
#
# This file is part of the The SCORE Framework.
#
# The SCORE Framework and all its parts are free software: you can redistribute
# them and/or modify them under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation which is in the
# file named COPYING.LESSER.txt.
#
# The SCORE Framework and all its parts are distributed without any WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. For more details see the GNU Lesser General Public
# License.
#
# If you have not received a copy of the GNU Lesser General Public License see
# http://www.gnu.org/licenses/.
#
# The License-Agreement realised between you as Licensee and STRG.AT GmbH as
# Licenser including the issue of its valid conclusion and its pre- and
# post-contractual effects is governed by the laws of Austria. Any disputes
# concerning this License-Agreement including the issue of its valid conclusion
# and its pre- and post-contractual effects are exclusively decided by the
# competent court, in whose district STRG.AT GmbH has its registered seat, at
# the discretion of STRG.AT GmbH also the competent court, in whose district the
# Licensee has his registered seat, an establishment or assets.

import os

from setuptools import setup

here = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(here, 'README.rst')) as f:
    README = f.read()

setup(
    name='score.serve',
    version='0.1.1',
    description='Application server of The SCORE Framework',
    long_description=README,
    author='strg.at',
    author_email='score@strg.at',
    url='http://score-framework.org',
    keywords='score framework server',
    packages=['score', 'score.serve', 'score.serve.worker'],
    namespace_packages=['score'],
    zip_safe=False,
    license='LGPL',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Console',
        'Environment :: Web Environment',
        'Framework :: Pyramid',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU Lesser General '
            'Public License v3 or later (LGPLv3+)',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Topic :: Software Development :: Libraries :: Application Frameworks',
    ],
    install_requires=[
        'score.init >= 0.3.9',
        'watchdog',
        'tblib',
    ],
    entry_points={
        'score.cli': [
            'serve = score.serve.cli:main',
        ]
    },
)



# Copyright © 2015,2016 STRG.AT GmbH, Vienna, Austria
#
# This file is part of the The SCORE Framework.
#
# The SCORE Framework and all its parts are free software: you can redistribute
# them and/or modify them under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation which is in the
# file named COPYING.LESSER.txt.
#
# The SCORE Framework and all its parts are distributed without any WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE. For more details see the GNU Lesser General Public
# License.
#
# If you have not received a copy of the GNU Lesser General Public License see
# http://www.gnu.org/licenses/.
#
# The License-Agreement realised between you as Licensee and STRG.AT GmbH as
# Licenser including the issue of its valid conclusion and its pre- and
# post-contractual effects is governed by the laws of Austria. Any disputes
# concerning this License-Agreement including the issue of its valid conclusion
# and its pre- and post-contractual effects are exclusively decided by the
# competent court, in whose district STRG.AT GmbH has its registered seat, at
# the discretion of STRG.AT GmbH also the competent court, in whose district the
# Licensee has his registered seat, an establishment or assets.

import click
from score.init import parse_config_file, init as score_init


@click.command('serve')
@click.pass_context
def main(clickctx):
    """
    Starts the application server
    """
    conf = parse_config_file(clickctx.obj['conf'].path)
    overrides = {
        'score.init': {
            'modules': 'score.serve',
        }
    }
    try:
        del conf['score.init']['autoimport']
    except KeyError:
        pass
    if 'serve' not in conf:
        conf['serve'] = {}
    if 'conf' not in conf['serve']:
        conf['serve']['conf'] = clickctx.obj['conf'].path
    score = score_init(conf, overrides=overrides)
    try:
        # delete logging configuration, we have already initialized logging
        # during score_init, above.
        del conf['formatters']
    except KeyError:
        pass
    score.serve.start()
